package com.mortals.xhx.base.framework.filter;

import cn.hutool.core.util.ObjectUtil;
import com.alibaba.fastjson.JSON;
import com.mortals.framework.common.Rest;
import com.mortals.framework.model.AccessLogPdu;
import com.mortals.framework.util.StringUtils;
import com.mortals.xhx.common.utils.IpUtils;
import com.mortals.xhx.system.MessageProducer;
import lombok.extern.slf4j.Slf4j;
import org.reactivestreams.Publisher;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cloud.gateway.filter.GatewayFilterChain;
import org.springframework.cloud.gateway.filter.GlobalFilter;
import org.springframework.cloud.gateway.filter.factory.rewrite.CachedBodyOutputMessage;
import org.springframework.cloud.gateway.route.Route;
import org.springframework.cloud.gateway.support.BodyInserterContext;
import org.springframework.cloud.gateway.support.ServerWebExchangeUtils;
import org.springframework.core.Ordered;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.core.io.buffer.DataBufferFactory;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.codec.HttpMessageReader;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.http.server.reactive.ServerHttpRequestDecorator;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.http.server.reactive.ServerHttpResponseDecorator;
import org.springframework.stereotype.Component;
import org.springframework.util.MultiValueMap;
import org.springframework.web.reactive.function.BodyInserter;
import org.springframework.web.reactive.function.BodyInserters;
import org.springframework.web.reactive.function.server.HandlerStrategies;
import org.springframework.web.reactive.function.server.ServerRequest;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * @author: zxfei
 * @date: 2022/8/18 23:52
 * @description:
 **/
@Slf4j
@Component
public class AccessLogFilter implements GlobalFilter, Ordered {
    @Value("${spring.application.name:gateway}")
    private String appName;
    @Autowired
    private MessageProducer messageProducer;

    private static final String TRACE_ID = "traceId";
    private static final String SPAN_ID = "spanId";
    private static final String PSPAN_ID = "pspanId";

    private final List<HttpMessageReader<?>> messageReaders = HandlerStrategies.withDefaults().messageReaders();


    /**
     * 顺序必须是<-1，否则标准的NettyWriteResponseFilter将在您的过滤器得到一个被调用的机会之前发送响应
     * 也就是说如果不小于 -1 ，将不会执行获取后端响应的逻辑
     *
     * @return
     */
    @Override
    public int getOrder() {
        return -2;
    }

    @Override
    @SuppressWarnings("unchecked")
    public Mono<Void> filter(ServerWebExchange exchange, GatewayFilterChain chain) {

        ServerHttpRequest request = exchange.getRequest();
        // 请求路径
        String requestPath = request.getPath().pathWithinApplication().value();

        Route route = getGatewayRoute(exchange);
        // String ipAddress = WebUtils.get(request);
        String ipAddress = IpUtils.getRealIpAddress(exchange.getRequest());
        Map<String, String> headers = exchange.getRequest().getHeaders().toSingleValueMap();
        AccessLogPdu accessLogPdu = new AccessLogPdu();
        accessLogPdu.initAttrValue();
        accessLogPdu.setAppName(appName);
        accessLogPdu.setTraceID(headers.getOrDefault(TRACE_ID, ""));
        accessLogPdu.setPspanId(Integer.parseInt(headers.getOrDefault(PSPAN_ID, "0")));
        accessLogPdu.setSpanId(Integer.parseInt(headers.getOrDefault(SPAN_ID, "1")));
        accessLogPdu.setSchemaData(request.getURI().getScheme());
        accessLogPdu.setHostName(request.getHeaders().getHost().getHostName());
        accessLogPdu.setUri(requestPath);
        accessLogPdu.setTargetServer(route.getId());
        accessLogPdu.setRequestIp(IpUtils.getRealIpAddress(exchange.getRequest()));
        accessLogPdu.setUa(headers.getOrDefault("User-Agent", ""));
        accessLogPdu.setRequestTime(new Date());
        accessLogPdu.setLogTime(new Date());
        accessLogPdu.setMethod(request.getMethodValue());

       // log.info("origin uri:{} host:{} port:{},contentLens", requestPath, request.getHeaders().getHost().getHostName(), request.getHeaders().getHost().getPort(),request.getHeaders().getContentLength());

        MediaType mediaType = request.getHeaders().getContentType();
        if (request.getHeaders().getContentLength() > 1024 * 64) {
            return chain.filter(exchange);
        }

        if (MediaType.APPLICATION_FORM_URLENCODED.isCompatibleWith(mediaType) || MediaType.APPLICATION_JSON.isCompatibleWith(mediaType)) {
            return writeBodyLog(exchange, chain, accessLogPdu);
        } else {
            return writeBasicLog(exchange, chain, accessLogPdu);
        }
    }

    /**
     * form参数形式
     *
     * @param exchange
     * @param chain
     * @param accessLog
     * @return
     */
    private Mono<Void> writeBasicLog(ServerWebExchange exchange, GatewayFilterChain chain, AccessLogPdu accessLog) {
        StringBuilder builder = new StringBuilder();
        MultiValueMap<String, String> queryParams = exchange.getRequest().getQueryParams();
        for (Map.Entry<String, List<String>> entry : queryParams.entrySet()) {
            builder.append(entry.getKey()).append("=").append(StringUtils.join(entry.getValue(), ","));
        }
        accessLog.setRequestData(builder.toString());
        //获取响应体
        ServerHttpResponseDecorator decoratedResponse = recordResponseLog(exchange, accessLog);
        return chain.filter(exchange.mutate().response(decoratedResponse).build())
                .then(Mono.fromRunnable(() -> {
                    // 打印日志 发送
                    writeAccessLog(accessLog);
                }));
    }


    /**
     * 解决 request body 只能读取一次问题，
     * 参考: org.springframework.cloud.gateway.filter.factory.rewrite.ModifyRequestBodyGatewayFilterFactory
     *
     * @param exchange
     * @param chain
     * @param accessLogPdu
     * @return
     */
    @SuppressWarnings("unchecked")
    private Mono writeBodyLog(ServerWebExchange exchange, GatewayFilterChain chain, AccessLogPdu accessLogPdu) {
        ServerRequest serverRequest = ServerRequest.create(exchange, messageReaders);
        Mono<String> modifiedBody = serverRequest.bodyToMono(String.class)
                .flatMap(body -> {
                    accessLogPdu.setRequestData(body);
                    return Mono.just(body);
                });

        // 通过 BodyInserter 插入 body(支持修改body), 避免 request body 只能获取一次
        BodyInserter bodyInserter = BodyInserters.fromPublisher(modifiedBody, String.class);
        HttpHeaders headers = new HttpHeaders();
        headers.putAll(exchange.getRequest().getHeaders());
        headers.remove(HttpHeaders.CONTENT_LENGTH);
        CachedBodyOutputMessage outputMessage = new CachedBodyOutputMessage(exchange, headers);

        return bodyInserter.insert(outputMessage, new BodyInserterContext())
                .then(Mono.defer(() -> {
                    // 重新封装请求
                    ServerHttpRequest decoratedRequest = requestDecorate(exchange, headers, outputMessage);
                    // 记录响应日志
                    ServerHttpResponseDecorator decoratedResponse = recordResponseLog(exchange, accessLogPdu);
                    // 记录普通的
                    return chain.filter(exchange.mutate().request(decoratedRequest).response(decoratedResponse).build())
                            .then(Mono.fromRunnable(() -> {
                                // 打印日志 并发送
                                writeAccessLog(accessLogPdu);
                            }));
                }));
    }

    private void writeAccessLog(AccessLogPdu accessLogPdu) {
        //messageProducer.syncAccessSend(accessLogPdu);
    }


    private Route getGatewayRoute(ServerWebExchange exchange) {
        return exchange.getAttribute(ServerWebExchangeUtils.GATEWAY_ROUTE_ATTR);
    }


    /**
     * 请求装饰器，重新计算 headers
     *
     * @param exchange
     * @param headers
     * @param outputMessage
     * @return
     */
    private ServerHttpRequestDecorator requestDecorate(ServerWebExchange exchange, HttpHeaders headers,
                                                       CachedBodyOutputMessage outputMessage) {
        return new ServerHttpRequestDecorator(exchange.getRequest()) {
            @Override
            public HttpHeaders getHeaders() {
                long contentLength = headers.getContentLength();
                log.debug("decoratedRequest contentLength:" + contentLength);
                HttpHeaders httpHeaders = new HttpHeaders();
                httpHeaders.putAll(super.getHeaders());
                if (contentLength > 0) {
                    httpHeaders.setContentLength(contentLength);
                } else {
                    httpHeaders.set(HttpHeaders.TRANSFER_ENCODING, "chunked");
                }
                return httpHeaders;
            }

            @Override
            public Flux<DataBuffer> getBody() {
                return outputMessage.getBody();
            }
        };
    }


    /**
     * 记录响应日志
     * 通过 DataBufferFactory 解决响应体分段传输问题。
     */
    private ServerHttpResponseDecorator recordResponseLog(ServerWebExchange exchange, AccessLogPdu accessLogPdu) {
        ServerHttpResponse response = exchange.getResponse();
        DataBufferFactory bufferFactory = response.bufferFactory();
        return new ServerHttpResponseDecorator(response) {
            @Override
            public Mono<Void> writeWith(Publisher<? extends DataBuffer> body) {
                if (body instanceof Flux) {
                    Date responseTime = new Date();
                    // accessLogPdu.setResponseTime(responseTime);
                    // 计算执行时间
                    long executeTime = (responseTime.getTime() - accessLogPdu.getRequestTime().getTime());
                    accessLogPdu.setResponseTime(responseTime);
                    accessLogPdu.setDuration(executeTime);
                    // 获取响应类型，如果是 json 就打印
                    String originalResponseContentType = exchange.getAttribute(ServerWebExchangeUtils.ORIGINAL_RESPONSE_CONTENT_TYPE_ATTR);
                    if (ObjectUtil.equal(this.getStatusCode(), HttpStatus.OK)
                            && StringUtils.isNotBlank(originalResponseContentType)) {
                        accessLogPdu.setResponseData(JSON.toJSONString(Rest.ok()));
                     /*   Flux<? extends DataBuffer> fluxBody = Flux.from(body);
                        return super.writeWith(fluxBody.buffer().map(dataBuffers -> {
                            // 合并多个流集合，解决返回体分段传输
                            DataBufferFactory dataBufferFactory = new DefaultDataBufferFactory();
                            DataBuffer join = dataBufferFactory.join(dataBuffers);
                            byte[] content = new byte[join.readableByteCount()];
                            join.read(content);
                            // 释放掉内存
                            DataBufferUtils.release(join);
                            accessLogPdu.setRequestData(JSON.toJSONString(Rest.ok()));
                            //String responseResult = new String(content, StandardCharsets.UTF_8);
                            //accessLogPdu.setResponseData(StrUtil.maxLength(responseResult,3000));
                            return bufferFactory.wrap(content);
                        }));*/
                    } else {
                        accessLogPdu.setResponseData(JSON.toJSONString(Rest.fail()));
                    }
                }
                // if body is not a flux. never got there.
                return super.writeWith(body);
            }
        };
    }
}
